
#include <iostream>

using namespace std;

#include <fstream>

#include "tm.h"
#include "aabb.h"
#include "m33.h"
#include "scene.h"

void TM::RotateAboutArbitraryAxis(V3 O, V3 a, float angled) {

	for (int vi = 0; vi < vertsN; vi++) {
		verts[vi] = verts[vi].RotateThisPointAboutArbitraryAxis(O, a, angled);
	}

}


void TM::RenderPoints(PPC *ppc, FrameBuffer *fb) {

	for (int vi = 0; vi < vertsN; vi++) {
		fb->Draw3DPoint(verts[vi], colors[vi], ppc, 7);
	}

}

void TM::RenderWireframe(PPC *ppc, FrameBuffer *fb) {

	for (int tri = 0; tri < trisN; tri++) {
		for (int ei = 0; ei < 3; ei++) {
			int vi0 = tris[3 * tri + ei];
			int vi1 = tris[3 * tri + ((ei + 1) % 3)];
			fb->Draw3DSegment(verts[vi0], colors[vi0], 
				verts[vi1], colors[vi1], ppc);
		}
	}

}

void TM::RenderFilled(PPC *ppc, FrameBuffer *fb) {

	if (!scene)
		return;

	V3 *pverts = new V3[vertsN];
	for (int vi = 0; vi < vertsN; vi++) {
		ppc->Project(verts[vi], pverts[vi]);
	}

	for (int tri = 0; tri < trisN; tri++) {
		unsigned int vis[3] = { tris[3 * tri + 0], tris[3 * tri + 1], tris[3 * tri + 2] };
		if (
			pverts[vis[0]][0] == FLT_MAX ||
			pverts[vis[1]][0] == FLT_MAX ||
			pverts[vis[2]][0] == FLT_MAX // implement clipping
			)
			continue;
		AABB aabb(pverts[vis[0]]);
		aabb.AddPoint(pverts[vis[1]]);
		aabb.AddPoint(pverts[vis[2]]);
		if (!aabb.Clip2D(0.0f, (float)fb->w, 0.0f, (float)fb->h))
			continue;
		int lefti = (int)(aabb.c0[0] + 0.5f);
		int righti = (int)(aabb.c1[0] - 0.5f);
		int topi= (int)(aabb.c0[1] + 0.5f);
		int bottomi = (int)(aabb.c1[1] - 0.5f);

		//setup edge equations
		M33 eeqs;
		for (int ei = 0; ei < 3; ei++) {
			V3 v0 = pverts[vis[ei]];
			V3 v1 = pverts[vis[(ei+1)%3]];
			V3 v2 = pverts[vis[(ei + 2) % 3]]; v2[2] = 1.0f;
			eeqs[ei][0] = v0[1] - v1[1];
			eeqs[ei][1] = v1[0] - v0[0];
			eeqs[ei][2] = -v0[0] * eeqs[ei][0] - v0[1] * eeqs[ei][1];
			if (eeqs[ei] * v2 < 0.0f)
				eeqs[ei] = eeqs[ei] * -1.0f;
		}

		M33 ssi;
		ssi[0] = pverts[vis[0]];
		ssi[1] = pverts[vis[1]];
		ssi[2] = pverts[vis[2]];
		V3 pvzs = ssi.GetColumn(2);
		ssi.SetColumn(2, V3(1.0f, 1.0f, 1.0f));
		ssi = ssi.Inverted();

		V3 zABC = ssi * pvzs;
		M33 colsm;
		colsm[0] = colors[vis[0]];
		colsm[1] = colors[vis[1]];
		colsm[2] = colors[vis[2]];
		M33 colsABC = ssi * colsm;
		colsABC = colsABC.Transposed();

		M33 normsm;
		normsm[0] = normals[vis[0]];
		normsm[1] = normals[vis[1]];
		normsm[2] = normals[vis[2]];
		M33 normsABC = ssi * normsm;
		normsABC = normsABC.Transposed();


		for (int v = topi; v <= bottomi; v++) {
			for (int u = lefti; u <= righti; u++) {
				V3 pc(.5f + (float)u, .5f + (float)v, 1.0f);
				V3 ss = eeqs * pc;
				if (ss[0] < 0.0f || ss[1] < 0.0f || ss[2] < 0.0f)
					continue;
				float currz = pc * zABC;
				if (!fb->Visible(u, v, currz))
					continue;
				V3 P = ppc->Unproject(V3(pc[0], pc[1], currz));
				V3 eyeRay = ppc->C - P;
				V3 currNorm = normsABC * pc; currNorm = currNorm.UnitVector();
				V3 refRay = currNorm.Reflect(eyeRay);
				V3 currCol = colsABC * pc;
				refRay = refRay.UnitVector();

				float ka = 0.2f;
				V3 lv = (scene->L - P).UnitVector();
				float kd = currNorm*lv;
				kd = (kd < 0.0f) ? 0.0f : kd;
				currCol = currCol * (ka + (1.0f - ka)*kd);

				float ks = lv*refRay;
				if (ks < 0.0f) ks = 0.0f;
				float kse = 200.0f;
				ks = pow(ks, kse);
				currCol = (currCol + V3(1.0f, 1.0f, 1.0f)*ks).Clamp();

//				currCol = (refRay + V3(1.0f, 1.0f, 1.0f)) * 0.5f;
				fb->Set(u, v, currCol.GetColor());
			}
		}
	}

}


void TM::SetRectangle(V3 O, float rw, float rh) {

	vertsN = 4;
	trisN = 2;
	Allocate();

	verts[0] = O + V3(-rw / 2.0f, +rh / 2.0f, 0.0f);
	verts[1] = O + V3(-rw / 2.0f, -rh / 2.0f, 0.0f);
	verts[2] = O + V3(+rw / 2.0f, -rh / 2.0f, 0.0f);
	verts[3] = O + V3(+rw / 2.0f, +rh / 2.0f, 0.0f);

	int tri = 0;
	tris[3 * tri + 0] = 0;
	tris[3 * tri + 1] = 1;
	tris[3 * tri + 2] = 2;
	tri++;

	tris[3 * tri + 0] = 2;
	tris[3 * tri + 1] = 3;
	tris[3 * tri + 2] = 0;
	tri++;

	for (int vi = 0; vi < vertsN; vi++) {
		colors[vi] = V3(0.0f, 0.0, 0.0f);
	}

}

void TM::Allocate() {

	verts = new V3[vertsN];
	colors = new V3[vertsN];
	tris = new unsigned int[3 * trisN];

}


void TM::LoadBin(char *fname) {

	ifstream ifs(fname, ios::binary);
	if (ifs.fail()) {
		cerr << "INFO: cannot open file: " << fname << endl;
		return;
	}

	ifs.read((char*)&vertsN, sizeof(int));
	char yn;
	ifs.read(&yn, 1); // always xyz
	if (yn != 'y') {
		cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
		return;
	}
	if (verts)
		delete verts;
	verts = new V3[vertsN];

	ifs.read(&yn, 1); // cols 3 floats
	if (colors)
		delete colors;
	colors = 0;
	if (yn == 'y') {
		colors = new V3[vertsN];
	}

	ifs.read(&yn, 1); // normals 3 floats
	if (normals)
		delete normals;
	normals = 0;
	if (yn == 'y') {
		normals = new V3[vertsN];
	}

	ifs.read(&yn, 1); // texture coordinates 2 floats
	float *tcs = 0; // don't have texture coordinates for now
	if (tcs)
		delete tcs;
	tcs = 0;
	if (yn == 'y') {
		tcs = new float[vertsN * 2];
	}

	ifs.read((char*)verts, vertsN * 3 * sizeof(float)); // load verts

	if (colors) {
		ifs.read((char*)colors, vertsN * 3 * sizeof(float)); // load cols
	}

	if (normals)
		ifs.read((char*)normals, vertsN * 3 * sizeof(float)); // load normals

	if (tcs)
		ifs.read((char*)tcs, vertsN * 2 * sizeof(float)); // load texture coordinates

	ifs.read((char*)&trisN, sizeof(int));
	if (tris)
		delete tris;
	tris = new unsigned int[trisN * 3];
	ifs.read((char*)tris, trisN * 3 * sizeof(unsigned int)); // read tiangles

	ifs.close();

	cerr << "INFO: loaded " << vertsN << " verts, " << trisN << " tris from " << endl << "      " << fname << endl;
	cerr << "      xyz " << ((colors) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;

	delete[]tcs;

}

AABB TM::GetAABB() {

	if (!verts)
		return AABB(V3(0.0f, 0.0f, 0.0f));

	AABB aabb(verts[0]);
	for (int i = 1; i < vertsN; i++)
		aabb.AddPoint(verts[i]);

	return aabb;

}

V3 TM::GetCenter() {

	V3 tmc(0.0f, 0.0f, 0.0f);
	for (int vi = 0; vi < vertsN; vi++) {

		tmc = tmc + verts[vi];

	}

	tmc = tmc * (1.0f / (float)vertsN);

	return tmc;

}

void TM::PositionAndSize(V3 tmC, float tmSize) {

	AABB aabb = GetAABB();
	V3 oldC = aabb.GetCenter();
	float oldSize = aabb.GetDiagonal();

	Translate(V3(0.0f, 0.0f, 0.0f) - oldC);
	Scale(tmSize / oldSize);
	Translate(tmC);

}

void TM::Translate(V3 tv) {

	for (int vi = 0; vi < vertsN; vi++)
		verts[vi] = verts[vi] + tv;

}

void TM::Scale(float scf) {

	for (int vi = 0; vi < vertsN; vi++)
		verts[vi] = verts[vi]*scf;

}

void TM::Light(V3 mc, V3 L) {

	for (int vi = 0; vi < vertsN; vi++) {
		float ka = 0.5f;
		float kd = (L - verts[vi]).UnitVector() * normals[vi].UnitVector();
		kd = (kd < 0.0f) ? 0.0f : kd;
		colors[vi] = mc * (ka + (1.0f - ka)*kd);
	}

}
